# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import hashlib
import itertools
import json
import os

import pytest

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.bindings._rust import openssl as rust_openssl
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, mlkem, x25519
from cryptography.hazmat.primitives.hpke import (
    AEAD,
    KDF,
    KEM,
    MLKEM768X25519PrivateKey,
    MLKEM768X25519PublicKey,
    MLKEM1024P384PrivateKey,
    MLKEM1024P384PublicKey,
    Suite,
)

from ...utils import load_vectors_from_file


def _hybrid_from_xwing_seed(
    seed: bytes,
) -> MLKEM768X25519PrivateKey:
    # X-Wing seed expansion: SHAKE256(seed, 96) -> (d || z || sk_X).
    expanded = hashlib.shake_256(seed).digest(96)
    mlkem_sk = mlkem.MLKEM768PrivateKey.from_seed_bytes(expanded[:64])
    x25519_sk = x25519.X25519PrivateKey.from_private_bytes(expanded[64:96])
    return MLKEM768X25519PrivateKey(mlkem_sk, x25519_sk)


def _hybrid_from_mlkem1024_p384_seed(
    seed: bytes,
) -> MLKEM1024P384PrivateKey:
    # MLKEM1024-P384 seed expansion: SHAKE256(seed, 112) -> (seed_PQ (64) ||
    # seed_T (48)).
    expanded = hashlib.shake_256(seed).digest(112)
    mlkem_sk = mlkem.MLKEM1024PrivateKey.from_seed_bytes(expanded[:64])
    p384_value = int.from_bytes(expanded[64:112], "big")
    p384_sk = ec.derive_private_key(p384_value, ec.SECP384R1())
    return MLKEM1024P384PrivateKey(mlkem_sk, p384_sk)


X25519_ENC_LENGTH = 32
P256_ENC_LENGTH = 65
P384_ENC_LENGTH = 97
P521_ENC_LENGTH = 133
MLKEM768_ENC_LENGTH = 1088
MLKEM1024_ENC_LENGTH = 1568
MLKEM768_X25519_ENC_LENGTH = 1120
MLKEM1024_P384_ENC_LENGTH = 1665

SUPPORTED_SUITES = list(
    itertools.product(
        [
            KEM.X25519,
            KEM.P256,
            KEM.P384,
            KEM.P521,
            KEM.MLKEM768,
            KEM.MLKEM1024,
            KEM.MLKEM768_X25519,
            KEM.MLKEM1024_P384,
        ],
        [
            KDF.HKDF_SHA256,
            KDF.HKDF_SHA384,
            KDF.HKDF_SHA512,
            KDF.SHAKE128,
            KDF.SHAKE256,
        ],
        [AEAD.AES_128_GCM, AEAD.AES_256_GCM, AEAD.CHACHA20_POLY1305],
    )
)


@pytest.mark.supported(
    only_if=lambda backend: backend.x25519_supported(),
    skip_message="Requires OpenSSL with X25519 support",
)
class TestHPKE:
    def test_invalid_kem_type(self):
        with pytest.raises(TypeError):
            Suite("not a kem", KDF.HKDF_SHA256, AEAD.AES_128_GCM)  # type: ignore[arg-type]

    def test_invalid_kdf_type(self):
        with pytest.raises(TypeError):
            Suite(KEM.X25519, "not a kdf", AEAD.AES_128_GCM)  # type: ignore[arg-type]

    def test_invalid_aead_type(self):
        with pytest.raises(TypeError):
            Suite(KEM.X25519, KDF.HKDF_SHA256, "not an aead")  # type: ignore[arg-type]

    @pytest.mark.parametrize(
        "kem,expected",
        [
            (KEM.X25519, X25519_ENC_LENGTH),
            (KEM.P256, P256_ENC_LENGTH),
            (KEM.P384, P384_ENC_LENGTH),
            (KEM.P521, P521_ENC_LENGTH),
            (KEM.MLKEM768, MLKEM768_ENC_LENGTH),
            (KEM.MLKEM1024, MLKEM1024_ENC_LENGTH),
            (KEM.MLKEM768_X25519, MLKEM768_X25519_ENC_LENGTH),
            (KEM.MLKEM1024_P384, MLKEM1024_P384_ENC_LENGTH),
        ],
    )
    def test_enc_length(self, kem, expected):
        assert kem.enc_length() == expected

    @pytest.mark.parametrize("kem,kdf,aead", SUPPORTED_SUITES)
    def test_roundtrip(self, backend, kem, kdf, aead):
        if kdf == KDF.SHAKE128 and not backend.hash_supported(
            hashes.SHAKE128(digest_size=32)
        ):
            pytest.skip("SHAKE128 not supported")
        if kdf == KDF.SHAKE256 and not backend.hash_supported(
            hashes.SHAKE256(digest_size=64)
        ):
            pytest.skip("SHAKE256 not supported")
        if (
            kem
            in [
                KEM.MLKEM768,
                KEM.MLKEM1024,
                KEM.MLKEM768_X25519,
                KEM.MLKEM1024_P384,
            ]
            and not backend.mlkem_supported()
        ):
            pytest.skip("ML-KEM not supported")
        if kem in [
            KEM.MLKEM768_X25519,
            KEM.MLKEM1024_P384,
        ] and not backend.hash_supported(hashes.SHA3_256()):
            pytest.skip("SHA3-256 not supported")
        suite = Suite(kem, kdf, aead)

        sk_r: (
            x25519.X25519PrivateKey
            | ec.EllipticCurvePrivateKey
            | mlkem.MLKEM768PrivateKey
            | mlkem.MLKEM1024PrivateKey
            | MLKEM768X25519PrivateKey
            | MLKEM1024P384PrivateKey
        )
        if kem == KEM.X25519:
            sk_r = x25519.X25519PrivateKey.generate()
        elif kem == KEM.P256:
            sk_r = ec.generate_private_key(ec.SECP256R1())
        elif kem == KEM.P384:
            sk_r = ec.generate_private_key(ec.SECP384R1())
        elif kem == KEM.P521:
            sk_r = ec.generate_private_key(ec.SECP521R1())
        elif kem == KEM.MLKEM768:
            sk_r = mlkem.MLKEM768PrivateKey.generate()
        elif kem == KEM.MLKEM1024:
            sk_r = mlkem.MLKEM1024PrivateKey.generate()
        elif kem == KEM.MLKEM768_X25519:
            sk_r = MLKEM768X25519PrivateKey(
                mlkem.MLKEM768PrivateKey.generate(),
                x25519.X25519PrivateKey.generate(),
            )
        else:
            sk_r = MLKEM1024P384PrivateKey(
                mlkem.MLKEM1024PrivateKey.generate(),
                ec.generate_private_key(ec.SECP384R1()),
            )
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"Hello, HPKE!", pk_r, info=b"test")
        plaintext = suite.decrypt(ciphertext, sk_r, info=b"test")

        assert plaintext == b"Hello, HPKE!"

    @pytest.mark.parametrize("kem,kdf,aead", SUPPORTED_SUITES)
    def test_roundtrip_no_info(self, backend, kem, kdf, aead):
        if kdf == KDF.SHAKE128 and not backend.hash_supported(
            hashes.SHAKE128(digest_size=32)
        ):
            pytest.skip("SHAKE128 not supported")
        if kdf == KDF.SHAKE256 and not backend.hash_supported(
            hashes.SHAKE256(digest_size=64)
        ):
            pytest.skip("SHAKE256 not supported")
        if (
            kem
            in [
                KEM.MLKEM768,
                KEM.MLKEM1024,
                KEM.MLKEM768_X25519,
                KEM.MLKEM1024_P384,
            ]
            and not backend.mlkem_supported()
        ):
            pytest.skip("ML-KEM not supported")
        if kem in [
            KEM.MLKEM768_X25519,
            KEM.MLKEM1024_P384,
        ] and not backend.hash_supported(hashes.SHA3_256()):
            pytest.skip("SHA3-256 not supported")
        suite = Suite(kem, kdf, aead)

        sk_r: (
            x25519.X25519PrivateKey
            | ec.EllipticCurvePrivateKey
            | mlkem.MLKEM768PrivateKey
            | mlkem.MLKEM1024PrivateKey
            | MLKEM768X25519PrivateKey
            | MLKEM1024P384PrivateKey
        )
        if kem == KEM.X25519:
            sk_r = x25519.X25519PrivateKey.generate()
        elif kem == KEM.P256:
            sk_r = ec.generate_private_key(ec.SECP256R1())
        elif kem == KEM.P384:
            sk_r = ec.generate_private_key(ec.SECP384R1())
        elif kem == KEM.P521:
            sk_r = ec.generate_private_key(ec.SECP521R1())
        elif kem == KEM.MLKEM768:
            sk_r = mlkem.MLKEM768PrivateKey.generate()
        elif kem == KEM.MLKEM1024:
            sk_r = mlkem.MLKEM1024PrivateKey.generate()
        elif kem == KEM.MLKEM768_X25519:
            sk_r = MLKEM768X25519PrivateKey(
                mlkem.MLKEM768PrivateKey.generate(),
                x25519.X25519PrivateKey.generate(),
            )
        else:
            sk_r = MLKEM1024P384PrivateKey(
                mlkem.MLKEM1024PrivateKey.generate(),
                ec.generate_private_key(ec.SECP384R1()),
            )
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"Hello!", pk_r)
        plaintext = suite.decrypt(ciphertext, sk_r)

        assert plaintext == b"Hello!"

    def test_wrong_key_x25519(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = x25519.X25519PrivateKey.generate()
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        ec_pk = ec.generate_private_key(ec.SECP256R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", ec_pk)

        # Wrong key type for decrypt
        ec_sk = ec.generate_private_key(ec.SECP256R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, ec_sk)

    def test_wrong_key_p256(self):
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = ec.generate_private_key(ec.SECP256R1())
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = ec.generate_private_key(ec.SECP256R1())
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", x25519_pk)

        # Wrong key type for decrypt
        x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, x25519_sk)

        # Wrong EC curve for encrypt
        secp384r1_pk = ec.generate_private_key(ec.SECP384R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", secp384r1_pk)

        # Wrong EC curve for decrypt
        secp384r1_sk = ec.generate_private_key(ec.SECP384R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, secp384r1_sk)

    def test_wrong_key_p384(self):
        suite = Suite(KEM.P384, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = ec.generate_private_key(ec.SECP384R1())
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = ec.generate_private_key(ec.SECP384R1())
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", x25519_pk)

        # Wrong key type for decrypt
        x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, x25519_sk)

        # Wrong EC curve for encrypt
        secp256r1_pk = ec.generate_private_key(ec.SECP256R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", secp256r1_pk)

        # Wrong EC curve for decrypt
        secp256r1_sk = ec.generate_private_key(ec.SECP256R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, secp256r1_sk)

        # Wrong EC curve (P-521) for encrypt
        secp521r1_pk = ec.generate_private_key(ec.SECP521R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", secp521r1_pk)

        # Wrong EC curve (P-521) for decrypt
        secp521r1_sk = ec.generate_private_key(ec.SECP521R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, secp521r1_sk)

    def test_wrong_key_p521(self):
        suite = Suite(KEM.P521, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = ec.generate_private_key(ec.SECP521R1())
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = ec.generate_private_key(ec.SECP521R1())
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", x25519_pk)

        # Wrong key type for decrypt
        x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, x25519_sk)

        # Wrong EC curve for encrypt
        secp256r1_pk = ec.generate_private_key(ec.SECP256R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", secp256r1_pk)

        # Wrong EC curve for decrypt
        secp256r1_sk = ec.generate_private_key(ec.SECP256R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, secp256r1_sk)

        # Wrong EC curve (P-384) for encrypt
        secp384r1_pk = ec.generate_private_key(ec.SECP384R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", secp384r1_pk)

        # Wrong EC curve (P-384) for decrypt
        secp384r1_sk = ec.generate_private_key(ec.SECP384R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, secp384r1_sk)

    def test_wrong_aad_fails(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = rust_openssl.hpke._encrypt_with_aad(
            suite, b"Secret message", pk_r, aad=b"correct aad"
        )

        with pytest.raises(InvalidTag):
            rust_openssl.hpke._decrypt_with_aad(
                suite, ciphertext, sk_r, aad=b"wrong aad"
            )

    def test_info_mismatch_fails(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"Secret", pk_r, info=b"sender info")

        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_r, info=b"different info")

    def test_ciphertext_format(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # ciphertext should be: enc (32 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == X25519_ENC_LENGTH + 4 + 16

    def test_ciphertext_format_p256(self):
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = ec.generate_private_key(ec.SECP256R1())
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # ciphertext should be: enc (65 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == P256_ENC_LENGTH + 4 + 16

    def test_ciphertext_format_p384(self):
        suite = Suite(KEM.P384, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = ec.generate_private_key(ec.SECP384R1())
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (97 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == P384_ENC_LENGTH + 4 + 16

    def test_ciphertext_format_p521(self):
        suite = Suite(KEM.P521, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = ec.generate_private_key(ec.SECP521R1())
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (133 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == P521_ENC_LENGTH + 4 + 16

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_ciphertext_format_mlkem768(self):
        suite = Suite(KEM.MLKEM768, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = mlkem.MLKEM768PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (1088 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == MLKEM768_ENC_LENGTH + 4 + 16

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_wrong_key_mlkem768(self):
        suite = Suite(KEM.MLKEM768, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = mlkem.MLKEM768PrivateKey.generate()
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = mlkem.MLKEM768PrivateKey.generate()
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", x25519_pk)

        # Wrong key type for decrypt
        x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, x25519_sk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem768_wrong_kem_with_ec(self):
        # ML-KEM public key with EC-based KEM suite should fail
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_pk = mlkem.MLKEM768PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", mlkem_pk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_ciphertext_format_mlkem1024(self):
        suite = Suite(KEM.MLKEM1024, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = mlkem.MLKEM1024PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (1568 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == MLKEM1024_ENC_LENGTH + 4 + 16

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_wrong_key_mlkem1024(self):
        suite = Suite(KEM.MLKEM1024, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = mlkem.MLKEM1024PrivateKey.generate()
        pk_r = sk_r.public_key()
        ciphertext = suite.encrypt(b"test", pk_r)

        # Wrong key of correct type
        sk_wrong = mlkem.MLKEM1024PrivateKey.generate()
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", x25519_pk)

        # Wrong key type for decrypt
        x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, x25519_sk)

        # ML-KEM-768 key with ML-KEM-1024 suite should fail
        mlkem768_pk = mlkem.MLKEM768PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", mlkem768_pk)

        mlkem768_sk = mlkem.MLKEM768PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, mlkem768_sk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem1024_wrong_kem_with_ec(self):
        # ML-KEM-1024 public key with EC-based KEM suite should fail
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_pk = mlkem.MLKEM1024PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", mlkem_pk)

    @pytest.mark.supported(
        only_if=lambda backend: (
            backend.mlkem_supported()
            and backend.hash_supported(hashes.SHA3_256())
        ),
        skip_message="Requires ML-KEM and SHA3-256 support",
    )
    def test_ciphertext_format_mlkem768_x25519(self):
        suite = Suite(KEM.MLKEM768_X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        mlkem_sk = mlkem.MLKEM768PrivateKey.generate()
        x25519_sk = x25519.X25519PrivateKey.generate()
        pk_r = MLKEM768X25519PublicKey(
            mlkem_sk.public_key(), x25519_sk.public_key()
        )

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (1120 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == MLKEM768_X25519_ENC_LENGTH + 4 + 16

    @pytest.mark.supported(
        only_if=lambda backend: (
            backend.mlkem_supported()
            and backend.hash_supported(hashes.SHA3_256())
        ),
        skip_message="Requires ML-KEM and SHA3-256 support",
    )
    def test_wrong_key_mlkem768_x25519(self):
        suite = Suite(KEM.MLKEM768_X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_sk = mlkem.MLKEM768PrivateKey.generate()
        x25519_sk = x25519.X25519PrivateKey.generate()
        sk_r = MLKEM768X25519PrivateKey(mlkem_sk, x25519_sk)
        pk_r = MLKEM768X25519PublicKey(
            mlkem_sk.public_key(), x25519_sk.public_key()
        )
        ciphertext = suite.encrypt(b"test", pk_r)
        # Correct key decrypts successfully.
        assert suite.decrypt(ciphertext, sk_r) == b"test"

        # Wrong key of correct type
        sk_wrong = MLKEM768X25519PrivateKey(
            mlkem.MLKEM768PrivateKey.generate(),
            x25519.X25519PrivateKey.generate(),
        )
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        stray_x25519_pk = x25519.X25519PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", stray_x25519_pk)

        # Wrong key type for decrypt
        stray_x25519_sk = x25519.X25519PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, stray_x25519_sk)

        # ML-KEM-768 key with hybrid suite should fail
        mlkem768_pk = mlkem.MLKEM768PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", mlkem768_pk)

        mlkem768_sk = mlkem.MLKEM768PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, mlkem768_sk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem768_x25519_wrong_kem_with_ec(self):
        # Hybrid public key with EC-based KEM suite should fail
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_sk = mlkem.MLKEM768PrivateKey.generate()
        x25519_sk = x25519.X25519PrivateKey.generate()
        hybrid_pk = MLKEM768X25519PublicKey(
            mlkem_sk.public_key(), x25519_sk.public_key()
        )
        with pytest.raises(TypeError):
            suite.encrypt(b"test", hybrid_pk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem768_x25519_constructor_type_errors(self):
        mlkem_sk = mlkem.MLKEM768PrivateKey.generate()
        mlkem_pk = mlkem_sk.public_key()
        x25519_sk = x25519.X25519PrivateKey.generate()
        x25519_pk = x25519_sk.public_key()

        # Wrong type for mlkem_key in private constructor.
        with pytest.raises(TypeError):
            MLKEM768X25519PrivateKey(x25519_sk, x25519_sk)  # type: ignore[arg-type]
        # Wrong type for x25519_key in private constructor.
        with pytest.raises(TypeError):
            MLKEM768X25519PrivateKey(mlkem_sk, mlkem_sk)  # type: ignore[arg-type]

        # Wrong type for mlkem_key in public constructor.
        with pytest.raises(TypeError):
            MLKEM768X25519PublicKey(x25519_pk, x25519_pk)  # type: ignore[arg-type]
        # Wrong type for x25519_key in public constructor.
        with pytest.raises(TypeError):
            MLKEM768X25519PublicKey(mlkem_pk, mlkem_pk)  # type: ignore[arg-type]

    @pytest.mark.supported(
        only_if=lambda backend: (
            backend.mlkem_supported()
            and backend.hash_supported(hashes.SHA3_256())
        ),
        skip_message="Requires ML-KEM and SHA3-256 support",
    )
    def test_ciphertext_format_mlkem1024_p384(self):
        suite = Suite(KEM.MLKEM1024_P384, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        mlkem_sk = mlkem.MLKEM1024PrivateKey.generate()
        p384_sk = ec.generate_private_key(ec.SECP384R1())
        pk_r = MLKEM1024P384PublicKey(
            mlkem_sk.public_key(), p384_sk.public_key()
        )

        ciphertext = suite.encrypt(b"test", pk_r)

        # enc (1665 bytes) + ct (4 bytes pt + 16 bytes tag)
        assert len(ciphertext) == MLKEM1024_P384_ENC_LENGTH + 4 + 16

    @pytest.mark.supported(
        only_if=lambda backend: (
            backend.mlkem_supported()
            and backend.hash_supported(hashes.SHA3_256())
        ),
        skip_message="Requires ML-KEM and SHA3-256 support",
    )
    def test_wrong_key_mlkem1024_p384(self):
        suite = Suite(KEM.MLKEM1024_P384, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_sk = mlkem.MLKEM1024PrivateKey.generate()
        p384_sk = ec.generate_private_key(ec.SECP384R1())
        sk_r = MLKEM1024P384PrivateKey(mlkem_sk, p384_sk)
        pk_r = MLKEM1024P384PublicKey(
            mlkem_sk.public_key(), p384_sk.public_key()
        )
        ciphertext = suite.encrypt(b"test", pk_r)
        # Correct key decrypts successfully.
        assert suite.decrypt(ciphertext, sk_r) == b"test"

        # Wrong key of correct type
        sk_wrong = MLKEM1024P384PrivateKey(
            mlkem.MLKEM1024PrivateKey.generate(),
            ec.generate_private_key(ec.SECP384R1()),
        )
        with pytest.raises(InvalidTag):
            suite.decrypt(ciphertext, sk_wrong)

        # Wrong key type for encrypt
        stray_p384_pk = ec.generate_private_key(ec.SECP384R1()).public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", stray_p384_pk)

        # Wrong key type for decrypt
        stray_p384_sk = ec.generate_private_key(ec.SECP384R1())
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, stray_p384_sk)

        # ML-KEM-1024 key with hybrid suite should fail
        mlkem1024_pk = mlkem.MLKEM1024PrivateKey.generate().public_key()
        with pytest.raises(TypeError):
            suite.encrypt(b"test", mlkem1024_pk)

        mlkem1024_sk = mlkem.MLKEM1024PrivateKey.generate()
        with pytest.raises(TypeError):
            suite.decrypt(ciphertext, mlkem1024_sk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem1024_p384_wrong_kem_with_ec(self):
        # Hybrid public key with EC-based KEM suite should fail
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        mlkem_sk = mlkem.MLKEM1024PrivateKey.generate()
        p384_sk = ec.generate_private_key(ec.SECP384R1())
        hybrid_pk = MLKEM1024P384PublicKey(
            mlkem_sk.public_key(), p384_sk.public_key()
        )
        with pytest.raises(TypeError):
            suite.encrypt(b"test", hybrid_pk)

    @pytest.mark.supported(
        only_if=lambda backend: backend.mlkem_supported(),
        skip_message="Requires ML-KEM support",
    )
    def test_mlkem1024_p384_constructor_type_errors(self):
        mlkem_sk = mlkem.MLKEM1024PrivateKey.generate()
        mlkem_pk = mlkem_sk.public_key()
        p384_sk = ec.generate_private_key(ec.SECP384R1())
        p384_pk = p384_sk.public_key()
        # Wrong-curve EC keys for curve validation paths.
        p256_sk = ec.generate_private_key(ec.SECP256R1())
        p256_pk = p256_sk.public_key()

        # Wrong type for mlkem_key in private constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PrivateKey(p384_sk, p384_sk)  # type: ignore[arg-type]
        # Wrong type for p384_key in private constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PrivateKey(mlkem_sk, mlkem_sk)  # type: ignore[arg-type]
        # Wrong EC curve for p384_key in private constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PrivateKey(mlkem_sk, p256_sk)

        # Wrong type for mlkem_key in public constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PublicKey(p384_pk, p384_pk)  # type: ignore[arg-type]
        # Wrong type for p384_key in public constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PublicKey(mlkem_pk, mlkem_pk)  # type: ignore[arg-type]
        # Wrong EC curve for p384_key in public constructor.
        with pytest.raises(TypeError):
            MLKEM1024P384PublicKey(mlkem_pk, p256_pk)

    def test_empty_plaintext(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"", pk_r)
        plaintext = suite.decrypt(ciphertext, sk_r)

        assert plaintext == b""

    def test_large_plaintext(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        large_message = b"A" * 100000

        ciphertext = suite.encrypt(large_message, pk_r)
        plaintext = suite.decrypt(ciphertext, sk_r)

        assert plaintext == large_message

    def test_corrupted_ciphertext(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # Corrupt the ciphertext (after enc)
        corrupted = (
            ciphertext[:X25519_ENC_LENGTH]
            + bytes([ciphertext[X25519_ENC_LENGTH] ^ 0xFF])
            + ciphertext[X25519_ENC_LENGTH + 1 :]
        )

        with pytest.raises(InvalidTag):
            suite.decrypt(corrupted, sk_r)

    def test_truncated_ciphertext(self):
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)

        sk_r = x25519.X25519PrivateKey.generate()
        pk_r = sk_r.public_key()

        ciphertext = suite.encrypt(b"test", pk_r)

        # Truncate the ciphertext
        truncated = ciphertext[:-1]

        with pytest.raises(InvalidTag):
            suite.decrypt(truncated, sk_r)
        with pytest.raises(InvalidTag):
            suite.decrypt(b"\x00", sk_r)

    @pytest.mark.parametrize(
        "small_order_point",
        [
            # All 8 known small-order points on Curve25519
            bytes(32),  # Zero point (order 1)
            bytes([1] + [0] * 31),  # Order 4
            bytes.fromhex(
                "ecffffffffffffffffffffffffffffffffffff"
                "ffffffffffffffffffffffffffff7f"
            ),  # Order 8
            bytes.fromhex(
                "5f9c95bca3508c24b1d0b1559c83ef5b04445cc"
                "4581c8e86d8224eddd09f1157"
            ),  # Order 8
            bytes.fromhex(
                "e0eb7a7c3b41b8ae1656e3faf19fc46ada098deb"
                "9c32b1fd866205165f49b800"
            ),  # Order 2
            bytes.fromhex(
                "0000000000000000000000000000000000000000"
                "000000000000000000000080"
            ),  # p (order 1)
            bytes.fromhex(
                "0100000000000000000000000000000000000000"
                "000000000000000000000080"
            ),  # p+1 (order 4)
            bytes.fromhex(
                "edffffffffffffffffffffffffffffffffffff"
                "ffffffffffffffffffffffffffff"
            ),  # p-1 (order 1)
        ],
    )
    def test_small_order_enc_raises_invalid_tag(self, small_order_point):
        """Small-order X25519 enc points must raise InvalidTag,
        not ValueError, to avoid leaking an error oracle."""
        suite = Suite(KEM.X25519, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = x25519.X25519PrivateKey.generate()

        # Build a fake ciphertext: small-order enc (32 bytes) + fake ct
        fake_ciphertext = small_order_point + b"\x00" * 32

        with pytest.raises(InvalidTag):
            suite.decrypt(fake_ciphertext, sk_r)

    def test_invalid_p256_enc_raises_invalid_tag(self):
        """Invalid P-256 enc points must raise InvalidTag."""
        suite = Suite(KEM.P256, KDF.HKDF_SHA256, AEAD.AES_128_GCM)
        sk_r = ec.generate_private_key(ec.SECP256R1())

        # Build a fake ciphertext: invalid enc (65 bytes) + fake ct
        fake_ciphertext = b"\x04" + b"\x00" * 64 + b"\x00" * 32

        with pytest.raises(InvalidTag):
            suite.decrypt(fake_ciphertext, sk_r)

    def test_info_too_large_fails_shake128(self, backend):
        if not backend.hash_supported(hashes.SHAKE128(digest_size=32)):
            pytest.skip("SHAKE128 not supported")
        suite = Suite(KEM.X25519, KDF.SHAKE128, AEAD.AES_128_GCM)
        pk_r = x25519.X25519PrivateKey.generate().public_key()

        with pytest.raises(ValueError, match="info is too large"):
            suite.encrypt(b"test", pk_r, info=b"x" * 65536)

    def test_info_too_large_fails_shake256(self, backend):
        if not backend.hash_supported(hashes.SHAKE256(digest_size=64)):
            pytest.skip("SHAKE256 not supported")
        suite = Suite(KEM.X25519, KDF.SHAKE256, AEAD.AES_128_GCM)
        pk_r = x25519.X25519PrivateKey.generate().public_key()

        with pytest.raises(ValueError, match="info is too large"):
            suite.encrypt(b"test", pk_r, info=b"x" * 65536)

    def test_vector_decryption(self, backend, subtests):
        rfc_vectors = load_vectors_from_file(
            os.path.join("HPKE", "test-vectors.json"),
            lambda f: json.load(f),
        )
        pq_vectors = load_vectors_from_file(
            os.path.join("HPKE", "hpke-pq-test-vectors.json"),
            lambda f: json.load(f),
        )

        kem_map = {
            0x0010: KEM.P256,
            0x0011: KEM.P384,
            0x0012: KEM.P521,
            0x0020: KEM.X25519,
            0x0041: KEM.MLKEM768,
            0x0042: KEM.MLKEM1024,
            0x0051: KEM.MLKEM1024_P384,
            0x647A: KEM.MLKEM768_X25519,
        }
        kdf_map = {
            0x0001: KDF.HKDF_SHA256,
            0x0002: KDF.HKDF_SHA384,
            0x0003: KDF.HKDF_SHA512,
            0x0010: KDF.SHAKE128,
            0x0011: KDF.SHAKE256,
        }
        aead_map = {
            0x0001: AEAD.AES_128_GCM,
            0x0002: AEAD.AES_256_GCM,
            0x0003: AEAD.CHACHA20_POLY1305,
        }

        for vector in rfc_vectors + pq_vectors:
            if not (
                vector["mode"] == 0
                and vector["kem_id"] in kem_map
                and vector["kdf_id"] in kdf_map
                and vector["aead_id"] in aead_map
            ):
                continue

            with subtests.test():
                kem = kem_map[vector["kem_id"]]
                kdf = kdf_map[vector["kdf_id"]]
                aead = aead_map[vector["aead_id"]]

                if kdf == KDF.SHAKE128 and not backend.hash_supported(
                    hashes.SHAKE128(digest_size=32)
                ):
                    continue
                if kdf == KDF.SHAKE256 and not backend.hash_supported(
                    hashes.SHAKE256(digest_size=64)
                ):
                    continue
                if (
                    kem
                    in [
                        KEM.MLKEM768,
                        KEM.MLKEM1024,
                        KEM.MLKEM768_X25519,
                        KEM.MLKEM1024_P384,
                    ]
                    and not backend.mlkem_supported()
                ):
                    continue
                if kem in [
                    KEM.MLKEM768_X25519,
                    KEM.MLKEM1024_P384,
                ] and not backend.hash_supported(hashes.SHA3_256()):
                    continue

                suite = Suite(kem, kdf, aead)

                sk_r_bytes = bytes.fromhex(vector["skRm"])
                sk_r: (
                    x25519.X25519PrivateKey
                    | ec.EllipticCurvePrivateKey
                    | mlkem.MLKEM768PrivateKey
                    | mlkem.MLKEM1024PrivateKey
                    | MLKEM768X25519PrivateKey
                    | MLKEM1024P384PrivateKey
                )
                if kem == KEM.X25519:
                    sk_r = x25519.X25519PrivateKey.from_private_bytes(
                        sk_r_bytes
                    )
                elif kem == KEM.P256:
                    private_value = int.from_bytes(sk_r_bytes, "big")
                    sk_r = ec.derive_private_key(private_value, ec.SECP256R1())
                elif kem == KEM.P384:
                    private_value = int.from_bytes(sk_r_bytes, "big")
                    sk_r = ec.derive_private_key(private_value, ec.SECP384R1())
                elif kem == KEM.P521:
                    private_value = int.from_bytes(sk_r_bytes, "big")
                    sk_r = ec.derive_private_key(private_value, ec.SECP521R1())
                elif kem == KEM.MLKEM768:
                    sk_r = mlkem.MLKEM768PrivateKey.from_seed_bytes(sk_r_bytes)
                elif kem == KEM.MLKEM1024:
                    sk_r = mlkem.MLKEM1024PrivateKey.from_seed_bytes(
                        sk_r_bytes
                    )
                elif kem == KEM.MLKEM768_X25519:
                    sk_r = _hybrid_from_xwing_seed(sk_r_bytes)
                else:
                    sk_r = _hybrid_from_mlkem1024_p384_seed(sk_r_bytes)

                enc = bytes.fromhex(vector["enc"])
                info = bytes.fromhex(vector["info"])

                # Test first encryption only (single-shot API)
                encryption = vector["encryptions"][0]
                aad = bytes.fromhex(encryption["aad"])
                ct = bytes.fromhex(encryption["ct"])
                pt_expected = bytes.fromhex(encryption["pt"])

                # Combine enc || ct for single-shot decrypt
                # Use internal function with AAD for test vector
                # validation
                ciphertext = enc + ct
                pt = rust_openssl.hpke._decrypt_with_aad(
                    suite, ciphertext, sk_r, info=info, aad=aad
                )
                assert pt == pt_expected



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

import glob
import itertools
import json
import os
import pathlib
import re
import sys
import uuid

import nox

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

nox.options.reuse_existing_virtualenvs = True
nox.options.default_venv_backend = "uv"


def install(
    session: nox.Session,
    *args: str,
    verbose: bool = True,
) -> None:
    if verbose:
        args += ("-v",)
    session.install(
        "-c",
        "ci-constraints-requirements.txt",
        *args,
        silent=False,
    )


def load_pyproject_toml() -> dict:
    with (pathlib.Path(__file__).parent / "pyproject.toml").open("rb") as f:
        return tomllib.load(f)


@nox.session
@nox.session(name="tests-ssh")
@nox.session(name="tests-randomorder")
@nox.session(name="tests-nocoverage")
@nox.session(name="tests-rust-debug")
def tests(session: nox.Session) -> None:
    extras = []
    groups = ["test"]
    if session.name == "tests-ssh":
        extras.append("ssh")
    if session.name == "tests-randomorder":
        groups = ["test-randomorder"]

    prof_location = (
        pathlib.Path(".") / ".rust-cov" / str(uuid.uuid4())
    ).absolute()
    if session.name != "tests-nocoverage":
        rustflags = os.environ.get("RUSTFLAGS", "")
        assert rustflags is not None
        session.env.update(
            {
                "RUSTFLAGS": f"-Cinstrument-coverage {rustflags}",
                "LLVM_PROFILE_FILE": str(prof_location / "cov-%p.profraw"),
            }
        )

    install_spec = f".[{','.join(extras)}]"
    install(session, "-e", "./vectors")
    if session.name == "tests-rust-debug":
        install(
            session,
            "--config-settings-package=cryptography:build-args=--profile=dev",
            install_spec,
        )
    else:
        install(session, install_spec)

    install(
        session,
        *itertools.chain.from_iterable(("--group", g) for g in groups),
    )

    session.run("uv", "pip", "list")

    if session.name != "tests-nocoverage":
        cov_args = [
            "--cov=cryptography",
            "--cov=tests",
        ]
    else:
        cov_args = []

    if session.posargs:
        tests = session.posargs
    else:
        tests = ["tests/"]

    session.run(
        "pytest",
        "-n",
        "auto",
        "--dist=worksteal",
        *cov_args,
        "--durations=10",
        *tests,
    )

    if session.name != "tests-nocoverage":
        [rust_so] = glob.glob(
            f"{session.virtualenv.location}/lib/**/cryptography/hazmat/bindings/_rust.*",
            recursive=True,
        )
        process_rust_coverage(session, [rust_so], prof_location)


@nox.session
def docs(session: nox.Session) -> None:
    install(session, ".[ssh]")
    install(
        session,
        "--group",
        "docs",
        "--group",
        "docstest",
        "--group",
        "sdist",
    )

    temp_dir = session.create_tmp()
    session.run(
        "sphinx-build",
        "-T",
        "-W",
        "-b",
        "html",
        "-d",
        f"{temp_dir}/doctrees",
        "docs",
        "docs/_build/html",
    )
    session.run(
        "sphinx-build",
        "-T",
        "-W",
        "-b",
        "latex",
        "-d",
        f"{temp_dir}/doctrees",
        "docs",
        "docs/_build/latex",
    )

    session.run(
        "sphinx-build",
        "-T",
        "-W",
        "-b",
        "doctest",
        "-d",
        f"{temp_dir}/doctrees",
        "docs",
        "docs/_build/html",
    )
    session.run(
        "sphinx-build",
        "-T",
        "-W",
        "-b",
        "spelling",
        "docs",
        "docs/_build/html",
    )

    session.run(
        "python3", "-m", "readme_renderer", "README.rst", "-o", "/dev/null"
    )
    session.run(
        "python3",
        "-m",
        "readme_renderer",
        "vectors/README.rst",
        "-o",
        "/dev/null",
    )


@nox.session(name="docs-linkcheck")
def docs_linkcheck(session: nox.Session) -> None:
    install(session, ".", "--group", "docs")

    session.run(
        "sphinx-build", "-W", "-b", "linkcheck", "docs", "docs/_build/html"
    )


@nox.session
def flake(session: nox.Session) -> None:
    # TODO: Ideally there'd be a pip flag to install just our dependencies,
    # but not install us.
    pyproject_data = load_pyproject_toml()
    install(session, "-e", "vectors/")
    install(
        session,
        *pyproject_data["build-system"]["requires"],
        *pyproject_data["project"]["optional-dependencies"]["ssh"],
        "--group",
        "pep8test",
        "--group",
        "test",
        "--group",
        "nox",
    )

    session.run("ruff", "check")
    session.run("ruff", "format", "--check")
    session.run(
        "mypy",
        "src/cryptography/",
        "vectors/cryptography_vectors/",
        "tests/",
        "release.py",
        "noxfile.py",
    )
    session.run("check-sdist", "--no-isolation")


@nox.session
@nox.session(name="rust-noclippy")
def rust(session: nox.Session) -> None:
    prof_location = (
        pathlib.Path(".") / ".rust-cov" / str(uuid.uuid4())
    ).absolute()
    rustflags = os.environ.get("RUSTFLAGS", "")
    assert rustflags is not None
    session.env.update(
        {
            "RUSTFLAGS": f"-Cinstrument-coverage  {rustflags}",
            "LLVM_PROFILE_FILE": str(prof_location / "cov-%p.profraw"),
        }
    )

    # TODO: Ideally there'd be a pip flag to install just our dependencies,
    # but not install us.
    pyproject_data = load_pyproject_toml()
    install(session, *pyproject_data["build-system"]["requires"])

    session.run("cargo", "fmt", "--all", "--", "--check", external=True)
    if session.name != "rust-noclippy":
        session.run(
            "cargo",
            "clippy",
            "--all",
            "--",
            "-D",
            "warnings",
            external=True,
        )

    build_output = session.run(
        "cargo",
        "test",
        "--all",
        "--no-run",
        "-q",
        "--message-format=json",
        external=True,
        silent=True,
    )
    session.run("cargo", "test", "--all", external=True)

    # It's None on install-only invocations
    if build_output is not None:
        assert isinstance(build_output, str)
        rust_tests = []
        for line in build_output.splitlines():
            data = json.loads(line)
            if data.get("profile", {}).get("test", False):
                rust_tests.extend(data["filenames"])

        process_rust_coverage(session, rust_tests, prof_location)


@nox.session
def local(session: nox.Session):
    if "CARGO_INCREMENTAL" not in os.environ:
        session.env["CARGO_INCREMENTAL"] = "1"

    pyproject_data = load_pyproject_toml()
    install(session, "-e", "./vectors", verbose=False)
    install(
        session,
        *pyproject_data["build-system"]["requires"],
        *pyproject_data["project"]["optional-dependencies"]["ssh"],
        *nox.project.dependency_groups(
            pyproject_data, "pep8test", "test", "nox"
        ),
        verbose=False,
    )

    session.run("ruff", "format")
    session.run("ruff", "check")

    session.run("cargo", "fmt", "--all", external=True)
    session.run("cargo", "check", "--all", "--tests", external=True)
    session.run(
        "cargo",
        "clippy",
        "--all",
        "--",
        "-D",
        "warnings",
        external=True,
    )

    session.run(
        "mypy",
        "src/cryptography/",
        "vectors/cryptography_vectors/",
        "tests/",
        "release.py",
        "noxfile.py",
    )

    session.run(
        "maturin",
        "develop",
        "--release",
        "--uv",
    )

    if session.posargs:
        tests = session.posargs
    else:
        tests = ["tests/"]

    session.run(
        "pytest",
        "-n",
        "auto",
        "--dist=worksteal",
        "--durations=10",
        *tests,
    )

    session.run("cargo", "test", "--all", external=True)


LCOV_SOURCEFILE_RE = re.compile(
    r"^SF:.*[\\/]src[\\/]rust[\\/](.*)$", flags=re.MULTILINE
)
BIN_EXT = ".exe" if sys.platform == "win32" else ""


def process_rust_coverage(
    session: nox.Session,
    rust_binaries: list[str],
    prof_raw_location: pathlib.Path,
) -> None:
    target_libdir = session.run(
        "rustc", "--print", "target-libdir", external=True, silent=True
    )
    if target_libdir is not None:
        target_bindir = pathlib.Path(target_libdir).parent / "bin"

        profraws = [
            str(prof_raw_location / p)
            for p in prof_raw_location.glob("*.profraw")
        ]
        session.run(
            str(target_bindir / ("llvm-profdata" + BIN_EXT)),
            "merge",
            "-sparse",
            *profraws,
            "-o",
            "rust-cov.profdata",
            external=True,
        )

        lcov_data = session.run(
            str(target_bindir / ("llvm-cov" + BIN_EXT)),
            "export",
            rust_binaries[0],
            *itertools.chain.from_iterable(
                ["-object", b] for b in rust_binaries[1:]
            ),
            "-instr-profile=rust-cov.profdata",
            "--ignore-filename-regex=[/\\].cargo[/\\]",
            "--ignore-filename-regex=[/\\]rustc[/\\]",
            "--ignore-filename-regex=[/\\].rustup[/\\]toolchains[/\\]",
            "--ignore-filename-regex=[/\\]target[/\\]",
            "--format=lcov",
            silent=True,
            external=True,
        )
        assert isinstance(lcov_data, str)
        lcov_data = LCOV_SOURCEFILE_RE.sub(
            lambda m: "SF:src/rust/" + m.group(1).replace("\\", "/"),
            lcov_data.replace("\r\n", "\n"),
        )
        with open(f"{uuid.uuid4()}.lcov", "w") as f:
            f.write(lcov_data)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

INCLUDES = """
#include <openssl/x509.h>

/*
 * See the comment above Cryptography_STACK_OF_X509 in x509.py
 */
typedef STACK_OF(X509_NAME) Cryptography_STACK_OF_X509_NAME;
"""

TYPES = """
typedef ... X509_NAME;
typedef ... X509_NAME_ENTRY;
typedef ... Cryptography_STACK_OF_X509_NAME;
"""

FUNCTIONS = """
void X509_NAME_free(X509_NAME *);

unsigned long X509_NAME_hash(X509_NAME *);

int i2d_X509_NAME(X509_NAME *, unsigned char **);
X509_NAME *d2i_X509_NAME(X509_NAME **, const unsigned char **, long);
X509_NAME_ENTRY *X509_NAME_delete_entry(X509_NAME *, int);
void X509_NAME_ENTRY_free(X509_NAME_ENTRY *);
int X509_NAME_get_index_by_NID(X509_NAME *, int, int);
int X509_NAME_cmp(const X509_NAME *, const X509_NAME *);
X509_NAME *X509_NAME_dup(X509_NAME *);
int X509_NAME_entry_count(const X509_NAME *);
X509_NAME_ENTRY *X509_NAME_get_entry(const X509_NAME *, int);
char *X509_NAME_oneline(const X509_NAME *, char *, int);

ASN1_OBJECT *X509_NAME_ENTRY_get_object(const X509_NAME_ENTRY *);
ASN1_STRING *X509_NAME_ENTRY_get_data(const X509_NAME_ENTRY *);

int X509_NAME_add_entry_by_NID(X509_NAME *, int, int, const unsigned char *,
                               int, int, int);

Cryptography_STACK_OF_X509_NAME *sk_X509_NAME_new_null(void);
int sk_X509_NAME_num(Cryptography_STACK_OF_X509_NAME *);
int sk_X509_NAME_push(Cryptography_STACK_OF_X509_NAME *, X509_NAME *);
X509_NAME *sk_X509_NAME_value(Cryptography_STACK_OF_X509_NAME *, int);
void sk_X509_NAME_free(Cryptography_STACK_OF_X509_NAME *);
"""

CUSTOMIZATIONS = """
"""



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

from cryptography import utils
from cryptography.exceptions import UnsupportedAlgorithm, _Reasons
from cryptography.hazmat.decrepit.ciphers.modes import CFB as CFB
from cryptography.hazmat.decrepit.ciphers.modes import CFB8 as CFB8
from cryptography.hazmat.decrepit.ciphers.modes import OFB as OFB
from cryptography.hazmat.primitives._cipheralgorithm import (
    BlockCipherAlgorithm,
    CipherAlgorithm,
)
from cryptography.hazmat.primitives._modes import (
    Mode as Mode,
)
from cryptography.hazmat.primitives._modes import (
    ModeWithAuthenticationTag as ModeWithAuthenticationTag,
)
from cryptography.hazmat.primitives._modes import (
    ModeWithInitializationVector as ModeWithInitializationVector,
)
from cryptography.hazmat.primitives._modes import (
    ModeWithNonce as ModeWithNonce,
)
from cryptography.hazmat.primitives._modes import (
    ModeWithTweak as ModeWithTweak,
)
from cryptography.hazmat.primitives._modes import (
    _check_aes_key_length,
    _check_iv_and_key_length,
    _check_nonce_length,
)
from cryptography.hazmat.primitives.ciphers import algorithms


class CBC(ModeWithInitializationVector):
    name = "CBC"

    def __init__(self, initialization_vector: utils.Buffer):
        utils._check_byteslike("initialization_vector", initialization_vector)
        self._initialization_vector = initialization_vector

    @property
    def initialization_vector(self) -> utils.Buffer:
        return self._initialization_vector

    validate_for_algorithm = _check_iv_and_key_length


class XTS(ModeWithTweak):
    name = "XTS"

    def __init__(self, tweak: utils.Buffer):
        utils._check_byteslike("tweak", tweak)

        if len(tweak) != 16:
            raise ValueError("tweak must be 128-bits (16 bytes)")

        self._tweak = tweak

    @property
    def tweak(self) -> utils.Buffer:
        return self._tweak

    def validate_for_algorithm(self, algorithm: CipherAlgorithm) -> None:
        if isinstance(algorithm, (algorithms.AES128, algorithms.AES256)):
            raise TypeError(
                "The AES128 and AES256 classes do not support XTS, please use "
                "the standard AES class instead."
            )

        if algorithm.key_size not in (256, 512):
            raise ValueError(
                "The XTS specification requires a 256-bit key for AES-128-XTS"
                " and 512-bit key for AES-256-XTS"
            )


class ECB(Mode):
    name = "ECB"

    validate_for_algorithm = _check_aes_key_length


class CTR(ModeWithNonce):
    name = "CTR"

    def __init__(self, nonce: utils.Buffer):
        utils._check_byteslike("nonce", nonce)
        self._nonce = nonce

    @property
    def nonce(self) -> utils.Buffer:
        return self._nonce

    def validate_for_algorithm(self, algorithm: CipherAlgorithm) -> None:
        _check_aes_key_length(self, algorithm)
        _check_nonce_length(self.nonce, self.name, algorithm)


class GCM(ModeWithInitializationVector, ModeWithAuthenticationTag):
    name = "GCM"
    _MAX_ENCRYPTED_BYTES = (2**39 - 256) // 8
    _MAX_AAD_BYTES = (2**64) // 8

    def __init__(
        self,
        initialization_vector: utils.Buffer,
        tag: bytes | None = None,
        min_tag_length: int = 16,
    ):
        # OpenSSL 3.0.0 constrains GCM IVs to [64, 1024] bits inclusive
        # This is a sane limit anyway so we'll enforce it here.
        utils._check_byteslike("initialization_vector", initialization_vector)
        if len(initialization_vector) < 8 or len(initialization_vector) > 128:
            raise ValueError(
                "initialization_vector must be between 8 and 128 bytes (64 "
                "and 1024 bits)."
            )
        self._initialization_vector = initialization_vector
        if min_tag_length < 4:
            raise ValueError("min_tag_length must be >= 4")
        if tag is not None:
            utils._check_bytes("tag", tag)
            if len(tag) < min_tag_length:
                raise ValueError(
                    f"Authentication tag must be {min_tag_length} bytes or "
                    "longer."
                )
        self._tag = tag
        self._min_tag_length = min_tag_length

    @property
    def tag(self) -> bytes | None:
        return self._tag

    @property
    def initialization_vector(self) -> utils.Buffer:
        return self._initialization_vector

    def validate_for_algorithm(self, algorithm: CipherAlgorithm) -> None:
        _check_aes_key_length(self, algorithm)
        if not isinstance(algorithm, BlockCipherAlgorithm):
            raise UnsupportedAlgorithm(
                "GCM requires a block cipher algorithm",
                _Reasons.UNSUPPORTED_CIPHER,
            )
        block_size_bytes = algorithm.block_size // 8
        if self._tag is not None and len(self._tag) > block_size_bytes:
            raise ValueError(
                f"Authentication tag cannot be more than {block_size_bytes} "
                "bytes."
            )


utils.deprecated(
    OFB,
    __name__,
    "OFB has been moved to "
    "cryptography.hazmat.decrepit.ciphers.modes.OFB and "
    "will be removed from "
    "cryptography.hazmat.primitives.ciphers.modes in 49.0.0.",
    utils.DeprecatedIn47,
    name="OFB",
)


utils.deprecated(
    CFB,
    __name__,
    "CFB has been moved to "
    "cryptography.hazmat.decrepit.ciphers.modes.CFB and "
    "will be removed from "
    "cryptography.hazmat.primitives.ciphers.modes in 49.0.0.",
    utils.DeprecatedIn47,
    name="CFB",
)


utils.deprecated(
    CFB8,
    __name__,
    "CFB8 has been moved to "
    "cryptography.hazmat.decrepit.ciphers.modes.CFB8 and "
    "will be removed from "
    "cryptography.hazmat.primitives.ciphers.modes in 49.0.0.",
    utils.DeprecatedIn47,
    name="CFB8",
)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from cryptography.hazmat.primitives.asymmetric import ec, mlkem, x25519
from cryptography.utils import Buffer

class KEM:
    X25519: KEM
    P256: KEM
    P384: KEM
    P521: KEM
    MLKEM768: KEM
    MLKEM1024: KEM
    MLKEM768_X25519: KEM
    MLKEM1024_P384: KEM
    def enc_length(self) -> int: ...

class KDF:
    HKDF_SHA256: KDF
    HKDF_SHA384: KDF
    HKDF_SHA512: KDF
    SHAKE128: KDF
    SHAKE256: KDF

class AEAD:
    AES_128_GCM: AEAD
    AES_256_GCM: AEAD
    CHACHA20_POLY1305: AEAD

class MLKEM768X25519PrivateKey:
    def __init__(
        self,
        mlkem_key: mlkem.MLKEM768PrivateKey,
        x25519_key: x25519.X25519PrivateKey,
    ) -> None: ...
    def public_key(self) -> MLKEM768X25519PublicKey: ...

class MLKEM768X25519PublicKey:
    def __init__(
        self,
        mlkem_key: mlkem.MLKEM768PublicKey,
        x25519_key: x25519.X25519PublicKey,
    ) -> None: ...

class MLKEM1024P384PrivateKey:
    def __init__(
        self,
        mlkem_key: mlkem.MLKEM1024PrivateKey,
        p384_key: ec.EllipticCurvePrivateKey,
    ) -> None: ...
    def public_key(self) -> MLKEM1024P384PublicKey: ...

class MLKEM1024P384PublicKey:
    def __init__(
        self,
        mlkem_key: mlkem.MLKEM1024PublicKey,
        p384_key: ec.EllipticCurvePublicKey,
    ) -> None: ...

class Suite:
    def __init__(self, kem: KEM, kdf: KDF, aead: AEAD) -> None: ...
    def encrypt(
        self,
        plaintext: Buffer,
        public_key: x25519.X25519PublicKey
        | ec.EllipticCurvePublicKey
        | mlkem.MLKEM768PublicKey
        | mlkem.MLKEM1024PublicKey
        | MLKEM768X25519PublicKey
        | MLKEM1024P384PublicKey,
        info: Buffer | None = None,
    ) -> bytes: ...
    def decrypt(
        self,
        ciphertext: Buffer,
        private_key: x25519.X25519PrivateKey
        | ec.EllipticCurvePrivateKey
        | mlkem.MLKEM768PrivateKey
        | mlkem.MLKEM1024PrivateKey
        | MLKEM768X25519PrivateKey
        | MLKEM1024P384PrivateKey,
        info: Buffer | None = None,
    ) -> bytes: ...

def _encrypt_with_aad(
    suite: Suite,
    plaintext: Buffer,
    public_key: x25519.X25519PublicKey
    | ec.EllipticCurvePublicKey
    | mlkem.MLKEM768PublicKey
    | mlkem.MLKEM1024PublicKey
    | MLKEM768X25519PublicKey
    | MLKEM1024P384PublicKey,
    info: Buffer | None = None,
    aad: Buffer | None = None,
) -> bytes: ...
def _decrypt_with_aad(
    suite: Suite,
    ciphertext: Buffer,
    private_key: x25519.X25519PrivateKey
    | ec.EllipticCurvePrivateKey
    | mlkem.MLKEM768PrivateKey
    | mlkem.MLKEM1024PrivateKey
    | MLKEM768X25519PrivateKey
    | MLKEM1024P384PrivateKey,
    info: Buffer | None = None,
    aad: Buffer | None = None,
) -> bytes: ...



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

INCLUDES = """
#include <openssl/bio.h>
"""

TYPES = """
typedef ... BIO;
typedef ... BIO_METHOD;
typedef ... BIO_ADDR;
"""

FUNCTIONS = """
int BIO_free(BIO *);
BIO *BIO_new_file(const char *, const char *);
int BIO_read(BIO *, void *, int);
int BIO_write(BIO *, const void *, int);

BIO *BIO_new(BIO_METHOD *);
const BIO_METHOD *BIO_s_mem(void);
BIO *BIO_new_mem_buf(const void *, int);
long BIO_set_mem_eof_return(BIO *, int);
long BIO_get_mem_data(BIO *, char **);
int BIO_should_read(BIO *);
int BIO_should_write(BIO *);
int BIO_should_io_special(BIO *);
int BIO_should_retry(BIO *);

BIO_ADDR *BIO_ADDR_new(void);
void BIO_ADDR_free(BIO_ADDR *);
"""

CUSTOMIZATIONS = """
#if CRYPTOGRAPHY_IS_LIBRESSL || CRYPTOGRAPHY_IS_BORINGSSL \
    || CRYPTOGRAPHY_IS_AWSLC

#if !defined(_WIN32)
#include <sys/socket.h>
#endif

#include <stdlib.h>
#if !CRYPTOGRAPHY_IS_AWSLC
typedef struct sockaddr BIO_ADDR;
#endif

BIO_ADDR *BIO_ADDR_new(void) {
    return malloc(sizeof(struct sockaddr_storage));
}

void BIO_ADDR_free(BIO_ADDR *ptr) {
    free(ptr);
}
#endif
"""



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import annotations

INCLUDES = """
#include <openssl/crypto.h>
"""

TYPES = """
static const int OPENSSL_VERSION;
static const int OPENSSL_CFLAGS;
static const int OPENSSL_BUILT_ON;
static const int OPENSSL_PLATFORM;
static const int OPENSSL_DIR;
"""

FUNCTIONS = """
void OPENSSL_cleanup(void);

unsigned long OpenSSL_version_num(void);
const char *OpenSSL_version(int);

void *OPENSSL_malloc(size_t);
void OPENSSL_free(void *);
"""

CUSTOMIZATIONS = """
"""



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.


import binascii
import sys

import pytest

from cryptography.exceptions import AlreadyFinalized, InvalidKey
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.concatkdf import (
    ConcatKDFHash,
    ConcatKDFHMAC,
)


class TestConcatKDFHash:
    def test_length_limit(self, backend):
        big_length = hashes.SHA256().digest_size * (2**32 - 1) + 1
        error = OverflowError if sys.maxsize <= 2**31 else ValueError

        with pytest.raises(error):
            ConcatKDFHash(hashes.SHA256(), big_length, None, backend)

    def test_already_finalized(self, backend):
        ckdf = ConcatKDFHash(hashes.SHA256(), 16, None, backend)

        ckdf.derive(b"\x01" * 16)

        with pytest.raises(AlreadyFinalized):
            ckdf.derive(b"\x02" * 16)

    def test_derive(self, backend):
        prk = binascii.unhexlify(
            b"52169af5c485dcc2321eb8d26d5efa21fb9b93c98e38412ee2484cf14f0d0d23"
        )

        okm = binascii.unhexlify(b"1c3bc9e7c4547c5191c0d478cccaed55")

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e53728157e634612c12d6d5223e204aeea4341565369647bd184bcd2"
            b"46f72971f292badaa2fe4124612cba"
        )

        ckdf = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)

        assert ckdf.derive(prk) == okm

    def test_buffer_protocol(self, backend):
        prk = binascii.unhexlify(
            b"52169af5c485dcc2321eb8d26d5efa21fb9b93c98e38412ee2484cf14f0d0d23"
        )

        okm = binascii.unhexlify(b"1c3bc9e7c4547c5191c0d478cccaed55")

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e53728157e634612c12d6d5223e204aeea4341565369647bd184bcd2"
            b"46f72971f292badaa2fe4124612cba"
        )

        ckdf = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)

        assert ckdf.derive(bytearray(prk)) == okm

    def test_verify(self, backend):
        prk = binascii.unhexlify(
            b"52169af5c485dcc2321eb8d26d5efa21fb9b93c98e38412ee2484cf14f0d0d23"
        )

        okm = binascii.unhexlify(b"1c3bc9e7c4547c5191c0d478cccaed55")

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e53728157e634612c12d6d5223e204aeea4341565369647bd184bcd2"
            b"46f72971f292badaa2fe4124612cba"
        )

        ckdf = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)

        ckdf.verify(prk, okm)

    def test_invalid_verify(self, backend):
        prk = binascii.unhexlify(
            b"52169af5c485dcc2321eb8d26d5efa21fb9b93c98e38412ee2484cf14f0d0d23"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e53728157e634612c12d6d5223e204aeea4341565369647bd184bcd2"
            b"46f72971f292badaa2fe4124612cba"
        )

        ckdf = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)

        with pytest.raises(InvalidKey):
            ckdf.verify(prk, b"wrong key")

    def test_unicode_typeerror(self, backend):
        with pytest.raises(TypeError):
            ConcatKDFHash(
                hashes.SHA256(),
                16,
                otherinfo="foo",  # type: ignore[arg-type]
                backend=backend,
            )

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHash(
                hashes.SHA256(), 16, otherinfo=None, backend=backend
            )

            ckdf.derive("foo")  # type: ignore[arg-type]

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHash(
                hashes.SHA256(), 16, otherinfo=None, backend=backend
            )

            ckdf.verify("foo", b"bar")  # type: ignore[arg-type]

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHash(
                hashes.SHA256(), 16, otherinfo=None, backend=backend
            )

            ckdf.verify(b"foo", "bar")  # type: ignore[arg-type]

    def test_derive_into(self, backend):
        prk = binascii.unhexlify(
            b"52169af5c485dcc2321eb8d26d5efa21fb9b93c98e38412ee2484cf14f0d0d23"
        )
        oinfo = binascii.unhexlify(
            b"a1b2c3d4e53728157e634612c12d6d5223e204aeea4341565369647bd184bcd2"
            b"46f72971f292badaa2fe4124612cba"
        )
        ckdf = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)
        buf = bytearray(16)
        n = ckdf.derive_into(prk, buf)
        assert n == 16
        # Verify the output matches what derive would produce
        ckdf2 = ConcatKDFHash(hashes.SHA256(), 16, oinfo, backend)
        expected = ckdf2.derive(prk)
        assert buf == expected

    @pytest.mark.parametrize(
        ("buflen", "outlen"), [(15, 16), (17, 16), (8, 16), (32, 16)]
    )
    def test_derive_into_buffer_incorrect_size(self, buflen, outlen, backend):
        ckdf = ConcatKDFHash(hashes.SHA256(), outlen, None, backend)
        buf = bytearray(buflen)
        with pytest.raises(ValueError, match="buffer must be"):
            ckdf.derive_into(b"key", buf)

    def test_derive_into_already_finalized(self, backend):
        ckdf = ConcatKDFHash(hashes.SHA256(), 16, None, backend)
        buf = bytearray(16)
        ckdf.derive_into(b"key", buf)
        with pytest.raises(AlreadyFinalized):
            ckdf.derive_into(b"key", buf)


class TestConcatKDFHMAC:
    def test_length_limit(self, backend):
        big_length = hashes.SHA256().digest_size * (2**32 - 1) + 1
        error = OverflowError if sys.maxsize <= 2**31 else ValueError

        with pytest.raises(error):
            ConcatKDFHMAC(hashes.SHA256(), big_length, None, None, backend)

    def test_already_finalized(self, backend):
        ckdf = ConcatKDFHMAC(hashes.SHA256(), 16, None, None, backend)

        ckdf.derive(b"\x01" * 16)

        with pytest.raises(AlreadyFinalized):
            ckdf.derive(b"\x02" * 16)

    def test_derive(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )

        okm = binascii.unhexlify(
            b"64ce901db10d558661f10b6836a122a7605323ce2f39bf27eaaac8b34cf89f2f"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )

        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)

        assert ckdf.derive(prk) == okm

    def test_buffer_protocol(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )

        okm = binascii.unhexlify(
            b"64ce901db10d558661f10b6836a122a7605323ce2f39bf27eaaac8b34cf89f2f"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )

        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)

        assert ckdf.derive(bytearray(prk)) == okm

    def test_derive_explicit_salt(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )

        okm = binascii.unhexlify(
            b"64ce901db10d558661f10b6836a122a7605323ce2f39bf27eaaac8b34cf89f2f"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )

        ckdf = ConcatKDFHMAC(
            hashes.SHA512(), 32, b"\x00" * 128, oinfo, backend
        )

        assert ckdf.derive(prk) == okm

    def test_verify(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )

        okm = binascii.unhexlify(
            b"64ce901db10d558661f10b6836a122a7605323ce2f39bf27eaaac8b34cf89f2f"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )

        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)

        ckdf.verify(prk, okm)

    def test_invalid_verify(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )

        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )

        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)

        with pytest.raises(InvalidKey):
            ckdf.verify(prk, b"wrong key")

    def test_unicode_typeerror(self, backend):
        with pytest.raises(TypeError):
            ConcatKDFHMAC(
                hashes.SHA256(),
                16,
                salt="foo",  # type: ignore[arg-type]
                otherinfo=None,
                backend=backend,
            )

        with pytest.raises(TypeError):
            ConcatKDFHMAC(
                hashes.SHA256(),
                16,
                salt=None,
                otherinfo="foo",  # type: ignore[arg-type]
                backend=backend,
            )

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHMAC(
                hashes.SHA256(), 16, salt=None, otherinfo=None, backend=backend
            )

            ckdf.derive("foo")  # type: ignore[arg-type]

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHMAC(
                hashes.SHA256(), 16, salt=None, otherinfo=None, backend=backend
            )

            ckdf.verify("foo", b"bar")  # type: ignore[arg-type]

        with pytest.raises(TypeError):
            ckdf = ConcatKDFHMAC(
                hashes.SHA256(), 16, salt=None, otherinfo=None, backend=backend
            )

            ckdf.verify(b"foo", "bar")  # type: ignore[arg-type]

    def test_unsupported_hash_algorithm(self, backend):
        # ConcatKDF requires a hash algorithm with an internal block size.
        with pytest.raises(TypeError):
            ConcatKDFHMAC(
                hashes.SHA3_256(),
                16,
                salt=None,
                otherinfo=None,
                backend=backend,
            )

    def test_derive_into(self, backend):
        prk = binascii.unhexlify(
            b"013951627c1dea63ea2d7702dd24e963eef5faac6b4af7e4"
            b"b831cde499dff1ce45f6179f741c728aa733583b02409208"
            b"8f0af7fce1d045edbc5790931e8d5ca79c73"
        )
        oinfo = binascii.unhexlify(
            b"a1b2c3d4e55e600be5f367e0e8a465f4bf2704db00c9325c"
            b"9fbd216d12b49160b2ae5157650f43415653696421e68e"
        )
        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)
        buf = bytearray(32)
        n = ckdf.derive_into(prk, buf)
        assert n == 32
        # Verify the output matches what derive would produce
        ckdf2 = ConcatKDFHMAC(hashes.SHA512(), 32, None, oinfo, backend)
        expected = ckdf2.derive(prk)
        assert buf == expected

    @pytest.mark.parametrize(
        ("buflen", "outlen"), [(31, 32), (33, 32), (16, 32), (64, 32)]
    )
    def test_derive_into_buffer_incorrect_size(self, buflen, outlen, backend):
        ckdf = ConcatKDFHMAC(hashes.SHA512(), outlen, None, None, backend)
        buf = bytearray(buflen)
        with pytest.raises(ValueError, match="buffer must be"):
            ckdf.derive_into(b"key", buf)

    def test_derive_into_already_finalized(self, backend):
        ckdf = ConcatKDFHMAC(hashes.SHA512(), 32, None, None, backend)
        buf = bytearray(32)
        ckdf.derive_into(b"key", buf)
        with pytest.raises(AlreadyFinalized):
            ckdf.derive_into(b"key", buf)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

# /// script
# dependencies = [
#     "click",
#     "packaging",
# ]
# ///

import pathlib
import re
import subprocess
import sys

import click
from packaging.version import Version

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def run(*args: str) -> None:
    print(f"[running] {list(args)}")
    subprocess.check_call(list(args))


@click.group()
def cli():
    pass


@cli.command()
def release() -> None:
    base_dir = pathlib.Path(__file__).parent
    with (base_dir / "pyproject.toml").open("rb") as f:
        pyproject = tomllib.load(f)
        version = pyproject["project"]["version"]

    if Version(version).is_prerelease:
        raise RuntimeError(
            f"Can't release, pyproject.toml version is pre-release: {version}"
        )

    # Tag and push the tag (this will trigger the wheel builder in Actions)
    run("git", "tag", "-s", version, "-m", f"{version} release")
    run("git", "push", "--tags", "git@github.com:pyca/cryptography.git")


def replace_pattern(p: pathlib.Path, pattern: str, replacement: str) -> None:
    content = p.read_text()
    match = re.search(pattern, content, re.MULTILINE)
    assert match is not None

    start, end = match.span()
    new_content = content[:start] + replacement + content[end:]
    p.write_text(new_content)


def replace_version(
    p: pathlib.Path, variable_name: str, new_version: str
) -> None:
    replace_pattern(
        p, rf"^{variable_name}\s*=\s*.*$", f'{variable_name} = "{new_version}"'
    )


def rust_version(version: Version) -> str:
    # The Rust crates use a version that maps to the cryptography version,
    # but pre-1.0: cryptography X.0.Z is version 0.X.Z of the crates, and
    # pre-releases map to semver pre-releases (49.0.0.dev1 is 0.49.0-dev1).
    if version.minor != 0:
        raise ValueError(
            f"Can't map {version} to a crate version, minor must be 0"
        )
    crate_version = f"0.{version.major}.{version.micro}"
    pre_parts = []
    if version.pre is not None:
        pre_parts.append(f"{version.pre[0]}{version.pre[1]}")
    if version.dev is not None:
        pre_parts.append(f"dev{version.dev}")
    if pre_parts:
        crate_version += "-" + ".".join(pre_parts)
    return crate_version


def bump_rust_versions(base_dir: pathlib.Path, new_version: str) -> None:
    # All crates in the workspace inherit the workspace version, so we only
    # need to update it in one place, then sync Cargo.lock.
    replace_version(
        base_dir / "Cargo.toml", "version", rust_version(Version(new_version))
    )
    run("cargo", "update", "--workspace")


@cli.command()
@click.argument("new_version")
def bump_version(new_version: str) -> None:
    base_dir = pathlib.Path(__file__).parent

    replace_version(base_dir / "pyproject.toml", "version", new_version)
    replace_version(
        base_dir / "src/cryptography/__about__.py", "__version__", new_version
    )
    replace_version(
        base_dir / "vectors/pyproject.toml",
        "version",
        new_version,
    )
    replace_version(
        base_dir / "vectors/cryptography_vectors/__about__.py",
        "__version__",
        new_version,
    )

    if Version(new_version).is_prerelease:
        replace_pattern(
            base_dir / "pyproject.toml",
            r'"cryptography_vectors(==.*?)?"',
            '"cryptography_vectors"',
        )
    else:
        replace_pattern(
            base_dir / "pyproject.toml",
            r'"cryptography_vectors(==.*?)?"',
            f'"cryptography_vectors=={new_version}"',
        )

    bump_rust_versions(base_dir, new_version)


if __name__ == "__main__":
    cli()
